/*******************************************************************************

  outer_control.c
  Questions, remarks and complaints may be directed to:  Reid Guenther
                                                         reid@npac.syr.edu

  This file contains an implementation of the outer control variates described
  in the Columbia paper, page 15 (general) and page 18 (specific).

*******************************************************************************/

#include <math.h>
#include <stdio.h>
#include "path.h"    /* should be included in mesh.h */
#include "option.h"  /* should be included in mesh.h */
#include "mesh.h"    /* mesh data structures */
#include "util.h"
#include "blackScholes.h"

/*
   Note: Q_hat[i] is just the mesh estimated option price from run i,
         u_hat[k][i] is the kth control variate computed given the
                     the meshData for run i,
         u[k] is the actual value that u_hat[k][i] is approximating.

   Note: I assume that the Number_of_Controls is either 1 or 2.  If more
         are desired then some memory allocation will be needed to
         compute all the different average combinations.
*/
double Compute_Outer_Control_Estimate(int Number_of_Controls, 
                                      int Number_of_Runs, double *Q_hat,
                                      double **u_hat, double *u)
{

   int i;
   double Avg_Q_hat, Avg_Q_hat_u1_hat, Avg_u1_hat, Avg_u1_hat_u2_hat;
   double Avg_u2_hat, Avg_Q_hat_u2_hat, Avg_u1_hat_sq, Avg_u2_hat_sq;
   double beta1, beta2, Q_bar_new, a11, a12, a21, a22;
   double a11_inv, a12_inv, a21_inv, a22_inv, rhs1, rhs2;
   double determinant_A;


/*
      compute the averages:
*/
      Avg_Q_hat = Avg_Q_hat_u1_hat = Avg_u1_hat = 0.0e0;
      for (i=0; i<Number_of_Runs; i++) 
      {
         Avg_Q_hat += Q_hat[i];
         Avg_u1_hat += u_hat[1][i];
         Avg_Q_hat_u1_hat += Q_hat[i]*u_hat[1][i];
         Avg_u1_hat_sq += u_hat[1][i]*u_hat[1][i];
      }

   switch (Number_of_Controls){

      case 1:
         beta1 = (Avg_Q_hat_u1_hat - Avg_Q_hat*Avg_u1_hat)/
                  (Avg_u1_hat_sq - Avg_u1_hat*Avg_u1_hat);

         Q_bar_new = Avg_Q_hat - beta1*(Avg_u1_hat - u[1]);

      break;
      case 2:
/* Compute the rest of the averages: */
         Avg_Q_hat_u2_hat = Avg_u1_hat_u2_hat = 0.0e0;
         Avg_u2_hat = Avg_u2_hat_sq = 0.0e0;
         for (i=0; i<Number_of_Runs; i++)
         {
            Avg_u2_hat += u_hat[2][i];
            Avg_u2_hat_sq += u_hat[2][i]*u_hat[2][i];
            Avg_Q_hat_u2_hat += Q_hat[i]*u_hat[2][i];
            Avg_u1_hat_u2_hat += u_hat[1][i]*u_hat[1][i];
         }
/* a11..a22 are elements of the 2x2 matrix that are part of the
  set of linear equations A[][]*beta[] = rhs[] */

         a11 = Avg_u1_hat_sq  - Avg_u1_hat*Avg_u1_hat;
         a12 = a21 = Avg_u1_hat_u2_hat  - Avg_u1_hat*Avg_u2_hat;
         a22 = Avg_u2_hat_sq  - Avg_u2_hat*Avg_u2_hat;
         rhs1 = Avg_Q_hat_u1_hat - Avg_Q_hat*Avg_u1_hat;
         rhs2 = Avg_Q_hat_u2_hat - Avg_Q_hat*Avg_u2_hat;
         determinant_A = a11*a22 - a12*a21;
         a11_inv =  a22/determinant_A;
         a12_inv = -a12/determinant_A;
         a21_inv = -a21/determinant_A;
         a22_inv =  a11/determinant_A;
         beta1 = rhs1*a11_inv + rhs2*a12_inv;
         beta2 = rhs1*a21_inv + rhs2*a22_inv;
         Q_bar_new = Avg_Q_hat - beta1*(Avg_u1_hat - u[1]) -
                              beta2*(Avg_u2_hat - u[2]);

      break;
   }

   return Q_bar_new;
      
}


/*******************************************************************************

   This routine is used to calculate the mesh estimated max european option
   value for the given final time option prices. 

*******************************************************************************/
double OuterCtlEuroMaxMeshValue(meshData *theMesh, double r,double  T,
                                double *option_t, double *option_tp1)
{
   int t, path_i, path_j, Final_t, underlier, Current_t;
   double max_current_price, *tmp;
   double max, option0;
   int trace;
   trace = 0;

   if (trace)
   {
      printf("OuterCtlMaxEuroOption>>> test \n");
      printf("OuterCtlMaxEuroOption>>> r=%lf \n",r);
      printf("OuterCtlMaxEuroOption>>> T=%lf \n",T);
   }
/*
  allocate the t+1 and t maximum european option price arrays:
*/
   Final_t = theMesh->option.pData.times;
/*
   option_t   = (double *)malloc(Final_t*sizeof(double));
   option_tp1 = (double *)malloc(Final_t*sizeof(double));

   option_t   = (double *)malloc(theMesh->meshPaths*sizeof(double));
   option_tp1 = (double *)malloc(theMesh->meshPaths*sizeof(double));

   allocate1dDouble(&option_t,   Final_t+1);
   allocate1dDouble(&option_tp1, Final_t+1);
*/

   if (trace)
   {
      printf("OuterCtlMaxEuroOption>>> test 2\n");
      printf("OuterCtlMaxEuroOption>>> paths: %i\n",theMesh->meshPaths);
      printf("OuterCtlMaxEuroOption>>> Final_t: %i\n",Final_t);
      printf("OuterCtlMaxEuroOption>>> underliers: %i\n",theMesh->option.pData.
              underliers);
      printf("OuterCtlMaxEuroOption>>> strike price: %lf\n",theMesh->option.
              strikePrice);
   }
/*
  First find the maximum at each mesh point at the final time
*/
   for (path_i=0; path_i< theMesh->meshPaths;path_i++)
   {
      max =  theMesh->option.pData.price[Final_t-1][path_i][0];
      for (underlier=1;underlier < theMesh->option.pData.underliers;
           underlier++)
      {
         if (max < theMesh->option.pData.price[Final_t-1][path_i][underlier])
             max = theMesh->option.pData.price[Final_t-1][path_i][underlier];
      }

/* this is a max call option */
      if (-theMesh->option.strikePrice + max > 0.0)
      {
         option_tp1[path_i] = -theMesh->option.strikePrice + max;
      }
      else
         option_tp1[path_i] = 0.0;

      option_t[path_i]   = 0.0;
   }

   if (trace)
   {
      printf("OuterCtlMaxEuroOption>>> test 3\n");
      for (path_i=0; path_i < theMesh->meshPaths; path_i++)
      {
         printf("OuterCtlMaxEuroOption>>> option_tp1[%i]=%lf\n",
                path_i, option_tp1[path_i]);
         for (path_j=0; path_j < theMesh->meshPaths; path_j++)
         {
            printf(
             "OuterCtlMaxEuroOption>>> theMesh->weights[%i][[%i][%i]=%lf\n",
                         Final_t-2,path_i,path_j,
                         theMesh->weights[Final_t-2][path_i][path_j]);
         }
      }
   }
      

/*
  Now calculate the mesh estimate of the European option value for the
  previously determined final option values.
*/
   for (Current_t = Final_t - 1; Current_t > 1; Current_t--)
   {
      for (path_i=0; path_i < theMesh->meshPaths; path_i++)
      {
         option_t[path_i] = 0.0;

         for (path_j=0; path_j < theMesh->meshPaths; path_j++)
         {
            option_t[path_i] += theMesh->weights[Current_t-1][path_i][path_j]*
                                option_tp1[path_j];
         }

         if (trace)
            printf("OuterCtlMaxEuroOption>>> option_t[%i]=%lf\n",
                path_i, option_t[path_i]);
      }
/*
   let option_t go to option_tp1
*/
      tmp = option_t;
      option_t = option_tp1;
      option_tp1 = tmp;
   }

   if (trace)
   {
      printf("OuterCtlMaxEuroOption>>> test 4\n");
   }
      
   Current_t = 0;
   path_i = 0;
   option_t[path_i] = 0.0;
   for (path_j=0; path_j < theMesh->meshPaths; path_j++)
   {
      option_t[path_i] += theMesh->weights[Current_t][path_i][path_j]*
                          option_tp1[path_j];
   }

   if (trace)
   {
      printf("OuterCtlMaxEuroOption>>> test 5\n");
   }

   option0 = option_t[path_i]*
             exp(-r*T);

   if (trace)
   {
      printf("OuterCtlMaxEuroOption>>> option value= %lf \n",
              option0);
   }
/*   free(option_t); 
   printf("I got to here first\n");
   free(option_tp1); 
   printf("I got to here\n"); */
   return (option0 > 0.0) ? option0 : 0.0;
}


void mecdf_(int *N_underliers,double *d,double *rho,
           double *probability,int *ier);
/*******************************************************************************

   This routine calculates the European option on the maximum of several assets.   using the method described in Herb Johnson's paper: 'Options on the Maximum
   or the Minimum of Several Assets,' Journal of Financial and Quantitative 
   Analysis, 23, 3, September 1988, 277-283.

   T = time that the option has to be exercised by (in units of years).
   r = risk free interest rate.

*******************************************************************************/
double EuropeanMaxOption(meshData *theMesh, double r, double v, double T, 
                         double *d, double *rho, double **rho2d)
{
   double EuroMaxOption_value;

   int    N_underliers, i, j, k, id_cur, irho_cur, ier;
   double S_i, S_j, X, sigma_i, sigma_j, sigma_k, sigma_ij, sigma_ik;
   double rho_ij, rho_jk, rho_ik, probability;
   int trace;
   trace = 0;

   N_underliers = theMesh->option.pData.underliers;
   X = theMesh->option.strikePrice;
	for (i=0;i<N_underliers;i++)
	{
		rho2d[i][i] = sqrt(v);
	}

	for (i=0;i<N_underliers;i++)
	{
		for (j=0;j<i;j++)
		{
			rho2d[i][j] = 0.0;
			rho2d[j][i] = 0.0;
		}
	}

   if (trace)
   {
/*      rho2d[0][1] = 0.5;
      rho2d[1][0] = 0.5;
      rho2d[0][0] = 0.3;
      rho2d[1][1] = 0.3;  */
      printf("EuroMaxOption>>> %i underliers\n",N_underliers);
      printf("EuroMaxOption>>> strike price: %lf\n",theMesh->option.
              strikePrice);
   }
/* d contains the probability vector, rho contains the upper triangular
   correlation coefficients for a particular multivariate cumulative normal
   distribution 
   d = (double *)malloc(N_underliers*sizeof(double));
   rho = (double *)malloc(N_underliers*(N_underliers-1)/2*sizeof(double));
*/
   EuroMaxOption_value = 0.0e0;
/* begin looping over the number of underliers to generate the positive term of
   the Johnson max option equation. */
   for (i=0; i<N_underliers; i++)
   {
      id_cur = irho_cur = 0;
      S_i = theMesh->option.pData.price[0][0][i];
      sigma_i = rho2d[i][i];
      d[id_cur] = (log(S_i/X) + (r + 
                   0.5e0*sigma_i*sigma_i)*T)/(sigma_i*sqrt(T));
      id_cur++;

      for (j=0; j<i; j++)
      {
         sigma_j = rho2d[j][j];
         S_j = theMesh->option.pData.price[0][0][j];
         rho_ij = rho2d[i][j];
         sigma_ij = sqrt(sigma_i*sigma_i - 2.0e0*rho_ij*sigma_i*sigma_j +
                    sigma_j*sigma_j);
         d[id_cur] = (log(S_i/S_j) + 0.5e0*sigma_ij*sigma_ij*T)/
                     (sigma_ij*sqrt(T));
         id_cur++;
      }

      for (j=i+1; j<N_underliers; j++)
      {
         sigma_j = rho2d[j][j];
         S_j = theMesh->option.pData.price[0][0][j];
         rho_ij = rho2d[i][j];
         sigma_ij = sqrt(sigma_i*sigma_i - 2.0e0*rho_ij*sigma_i*sigma_j +
                         sigma_j*sigma_j);
         d[id_cur] = (log(S_i/S_j) + 0.5e0*sigma_ij*sigma_ij*T)/
                     (sigma_ij*sqrt(T));
         id_cur++;
      }

/* now set up the correlation array: */

/* first fill the array with the ith row */
      for (k=0; k<i; k++)
      {
         sigma_j = sigma_i;
         sigma_k = rho2d[k][k];
         rho_ij = rho2d[i][i];
         rho_ik = rho2d[i][k];
         rho_jk = rho2d[i][k];
         sigma_ij = sqrt(sigma_i*sigma_i - 2.0e0*rho_ij*sigma_i*sigma_j +
                         sigma_j*sigma_j);
         sigma_ik = sqrt(sigma_i*sigma_i - 2.0e0*rho_ik*sigma_i*sigma_k +
                         sigma_k*sigma_k);
         rho[irho_cur] = (sigma_i - rho_ik*sigma_k)/sigma_ik;
/*         rho[irho_cur] = (sigma_i*sigma_i - rho_ij*sigma_i*sigma_j -
                          rho_ik*sigma_i*sigma_k + rho_jk*sigma_j*sigma_k)/
                         (sigma_ij*sigma_ik); */
         irho_cur++;
         if (trace)
         {
            printf("EuroMaxOption>>> 1 irho_cur=%i \n",irho_cur);
            printf("i = %d sigma_i = %lf\n", i,sigma_i);
            printf("j = %d sigma_j = %lf\n", j,sigma_j);
            printf("i = %d j = %d rho_ij = %lf\n", i,j,sigma_j);
            printf("i = %d j = %d sigma_ij = %lf\n", i,j,sigma_ij);
            printf("i = %d k = %d sigma_ij = %lf\n", i,k,sigma_ik);
            printf("Brute force sqrt(rho_ij)=%lf\n", sqrt(sigma_i*sigma_i - 
            2.0 * rho_ij * sigma_i *sigma_j + sigma_j*sigma_j));
            printf("Brute force 2 sqrt(rho_ik)=%lf\n", sqrt(sigma_i*sigma_i - 
            2.0 * rho_ik * sigma_i *sigma_k + sigma_k*sigma_k));
         }
      }

      for (k=i+1; k<N_underliers; k++)
      {
         sigma_j = sigma_i;
         sigma_k = rho2d[k][k];
         rho_ij = rho2d[i][i];
         rho_ik = rho2d[i][k];
         rho_jk = rho2d[i][k];
         sigma_ij = sqrt(sigma_i*sigma_i - 2.0e0*rho_ij*sigma_i*sigma_j +
                         sigma_j*sigma_j);
         sigma_ik = sqrt(sigma_i*sigma_i - 2.0e0*rho_ik*sigma_i*sigma_k +
                         sigma_k*sigma_k);
         rho[irho_cur] = (sigma_i - rho_ik*sigma_k)/sigma_ik;
/*         rho[irho_cur] = (sigma_i*sigma_i - rho_ij*sigma_i*sigma_j -
                          rho_ik*sigma_i*sigma_k + rho_jk*sigma_j*sigma_k)/
                         (sigma_ij*sigma_ik); */
         irho_cur++;
         if (trace)
         {
            printf("EuroMaxOption>>> 2 irho_cur=%i \n",irho_cur);
         }
      }

/* now fill the array with the rows above the ith */
      for (j=0;j<i;j++)
      {
         for (k=j+1;k<i;k++)
         {
            sigma_j = rho2d[j][j];
            sigma_k = rho2d[k][k];
            rho_ij = rho2d[i][j];
            rho_ik = rho2d[i][k];
            rho_jk = rho2d[j][k];
            sigma_ij = sqrt(sigma_i*sigma_i - 2.0e0*rho_ij*sigma_i*sigma_j +
                            sigma_j*sigma_j);
            sigma_ik = sqrt(sigma_i*sigma_i - 2.0e0*rho_ik*sigma_i*sigma_k +
                            sigma_k*sigma_k);
            rho[irho_cur] = (sigma_i*sigma_i - rho_ij*sigma_i*sigma_j -
                             rho_ik*sigma_i*sigma_k + rho_jk*sigma_j*sigma_k)/
                            (sigma_ij*sigma_ik);
            irho_cur++;
            if (trace)
            {
               printf("EuroMaxOption>>> 3 irho_cur=%i \n",irho_cur);
            }
         }

         for (k=i+1;k<N_underliers;k++)
         {
            sigma_j = rho2d[j][j];
            sigma_k = rho2d[k][k];
            rho_ij = rho2d[i][j];
            rho_ik = rho2d[i][k];
            rho_jk = rho2d[j][k];
            sigma_ij = sqrt(sigma_i*sigma_i - 2.0e0*rho_ij*sigma_i*sigma_j +
                            sigma_j*sigma_j);
            sigma_ik = sqrt(sigma_i*sigma_i - 2.0e0*rho_ik*sigma_i*sigma_k +
                            sigma_k*sigma_k);
            rho[irho_cur] = (sigma_i*sigma_i - rho_ij*sigma_i*sigma_j -
                             rho_ik*sigma_i*sigma_k + rho_jk*sigma_j*sigma_k)/
                            (sigma_ij*sigma_ik);
            irho_cur++;
            if (trace)
            {
               printf("EuroMaxOption>>> 4 irho_cur=%i \n",irho_cur);
            }
         }
      }

/* Then fill the array with the rows below the ith */
      for (j=i+1;j<N_underliers-1;j++)
      {
         for (k=j+1;k<N_underliers;k++)
         {
            sigma_j = rho2d[j][j];
            sigma_k = rho2d[k][k];
            rho_ij = rho2d[i][j];
            rho_ik = rho2d[i][k];
            rho_jk = rho2d[j][k];
            sigma_ij = sqrt(sigma_i*sigma_i - 2.0e0*rho_ij*sigma_i*sigma_j +
                            sigma_j*sigma_j);
            sigma_ik = sqrt(sigma_i*sigma_i - 2.0e0*rho_ik*sigma_i*sigma_k +
                            sigma_k*sigma_k);
            rho[irho_cur] = (sigma_i*sigma_i - rho_ij*sigma_i*sigma_j -
                             rho_ik*sigma_i*sigma_k + rho_jk*sigma_j*sigma_k)/
                            (sigma_ij*sigma_ik);
            irho_cur++;
            if (trace)
            {
               printf("EuroMaxOption>>> 5 irho_cur=%i \n",irho_cur);
            }
         }
      }
/* !!!!!!!!!!!!!! Have you changed mecdf so that d contains numbers such that
                  mecdf returns the probability that an array of random
                  normally distributed numbers are less than the d array ???? */
      printf("Ind value = %lf\n", BivarCND(d[0], d[1], rho[0]));
      printf("European>>> 1 d's are %lf %lf \n", d[0],d[1]);
      mecdf_(&N_underliers,d,rho,&probability,&ier);
      printf("European>>> 1 Cur value = %lf\n", probability);

      EuroMaxOption_value += S_i*probability;
      if (trace)
      {
         printf("ier=%i \n",ier);
         printf("EuroMaxOption>>> %i time, %lf stock, positive rhs prob:",
                i, S_i);
         printf(" %lf\n",probability);
         printf("\n");
         printf(" EuroMaxOption_value =%lf\n",probability);
      }
   }

/* Now loop over the number of underliers to calculate the negative part of the
   Johnson equation */
   id_cur = irho_cur = 0;
   for (i=0;i<N_underliers;i++)
   {
      S_i = theMesh->option.pData.price[0][0][i];
      sigma_i = rho2d[i][i];
      d[id_cur] = -(log(S_i/X) + (r - 0.5e0*sigma_i*sigma_i)*T)/
                   (sigma_i*sqrt(T));
      id_cur++;

      for (j=i+1;j<N_underliers;j++)
      {
         rho[irho_cur] = rho2d[i][j];
         irho_cur++;
         if (trace)
         {
            printf("EuroMaxOption>>> 7 irho_cur=%i \n",irho_cur);
         }
      }
   }
/* !!!!!!!!!!!!!! Have you changed mecdf so that d contains numbers such that
                  mecdf returns the probability that an array of random
                  normally distributed numbers are less than the d array ???? */
   printf("Ind value = %lf\n", BivarCND(d[0], d[1], rho[0]));
   mecdf_(&N_underliers,d,rho,&probability,&ier);
   EuroMaxOption_value -= X*exp(-r*T)*(1.0e0 - probability);
      
   if (trace)
   {
      printf("EuroMaxOption>>> %i time, %lf stock, negative rhs prob:",
             i, X);
      printf(" %lf\n",probability);
   }

/*   free(d);
   free(rho); */
   if (EuroMaxOption_value < 0.0)
      EuroMaxOption_value = 0.0;
   return EuroMaxOption_value;
}
